[data-md-color-scheme="rust-light"] {
  --md-primary-fg-color: #dea584;
  --md-primary-bg-color: #1f292d;
}

/* Override just a couple colors from slate */
[data-md-color-scheme="slate"] {
  --md-primary-fg-color: #dea584;
  --md-primary-bg-color: #1f292d;
}
